/*
 * File scanbase.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "tine.h"
#include "archive.h"

extern int nDaysInMonthTable[2][12];
#define nDaysInMonth(yy,mm) (((yy) % 4) ? nDaysInMonthTable[0][mm] : nDaysInMonthTable[1][mm])
 
int OpenDataFile(DAQRecord *daqr, int *hasAlternate);
int getNumLines(FILE *fp);
extern DAQRecord *daqTbl;
#define DATABLOB_SIZE 16384
BYTE g_datablob[DATABLOB_SIZE];
#define TRACE_BUFFER_SIZE DATABLOB_SIZE
BYTE *traceDataBuffer = g_datablob;
BYTE *targetDataBuffer = NULL;
int targetDataBufferSize = 0;
int totalNrRecords;
int nadbfiles = 0;
#define MAX_ADBFILES 20
char dbfile[64];
#define cctxt(txt) { printf("HDA: %s\n>",txt); ccerr(database_not_loaded); }
#define quit(i) { printf("HDA: %s\n>",erlst[i]); exit(1); }
int archdebug = 0;
int archPending = 0;
int outputlen = 5;
time_t startuptime;
time_t lastmonthtime;
int bailout = TRUE;
int dumpoutput = TRUE;
int scanonly = FALSE;
int addext = FALSE;
int usepoi = FALSE;
int g_isTerminating;
int AppendProblemsLog(char *s) { return 0;}
int nrecords;
int recordsCapacity;
int rmvrange = FALSE;
time_t rangestart = 0;
time_t rangestop = 0;
int pruneLength = 0;

DAQRecord *daqTbl;

#define addNumberFecs(f) {\
  sprintf(fn,"%s%s",ArchiveDBpath,f);\
  if ((fp = fopen(fn,"rt")) != NULL) {\
  totalNrRecords += getNumLines(fp);\
  fclose(fp); } }

int initScanner(void)
{
  int id,i,n,np,cc=0;
  static char fn[80],desc[256];
  char *c;
  FILE *fp;
 
  totalNrRecords = 0;
  addNumberFecs(dbfile);
  if (totalNrRecords == 0)
  {
    printf("no records found in database %s\n",dbfile);
    exit(1);
  }
  if ((daqTbl=calloc(totalNrRecords,sizeof(DAQRecord))) == NULL) ccerr(out_of_server_memory);
  if ((subSystems=(SUB_SYS *)calloc(totalNrRecords,sizeof(SUB_SYS))) == NULL) ccerr(out_of_server_memory);
  if ((prpAlias=(PRP_ALIAS *)calloc(totalNrRecords,sizeof(PRP_ALIAS))) == NULL) ccerr(out_of_server_memory);
 
  getFilterTable(); // this will sort out the 'useMiminalStorage business'
  if (!(n=loadDaqDb(dbfile,0))) cctxt("No Servers loaded!");
  if (!(np=loadDaqParamDb(n,0))) cctxt("No Parameters loaded!");
err:
  if (cc)
  {
    printf("error: %s\n>",erlst[cc]);
    if (daqTbl != NULL) free(daqTbl);
    exit(1);
  }
  return 0;
}
char *months[] = {"Jan","Feb","Mar","Apr","May","Jun",
                  "Jul","Aug","Sep","Oct","Nov","Dec"};
int getmonth(char *m)
{
  int i;
  for (i=0; i<12; i++) if (!stricmp(m,months[i])) return i+1;
  return 0;
}
int getEntryFromRectyp(int rec)
{
  int i;
 
  for (i=0; i<totalNrRecords; i++) if (daqTbl[i].rec == rec) return i;
  return -1;
}
int fileExists(int i,time_t *mtime)
{
  int hFile, found = 0;
  struct stat sbuf;
  daqTbl[i].hasLegacyStorage = FALSE;
  if ((hFile=OpenDataFile(&daqTbl[i],NULL)) != -1) 
  {
    found = -1;
    fstat(hFile,&sbuf);
    *mtime = sbuf.st_mtime;
    close(hFile);
  }
  return found;
}
int getArchivedEntryFromKeywordAndTime(char *keyword,int *rec,int *param,time_t tgt)
{
  DAQParameter *daqParam;
  int i,j,yy,mm,best=-1,keyExists=FALSE,activeId=-no_data_in_interval;
  time_t mtime = 0;
  struct tm *lt=localtime((time_t *)&tgt);

  yy = lt->tm_year; mm = lt->tm_mon + 1;
  for (i=0; i<totalNrRecords; i++)
  {
    if (daqTbl[i].option&NEVER) continue;
    daqParam = daqTbl[i].daqParam;
    for (j=0; j<daqTbl[i].nparam; j++)
    {
      if (!stricmp(daqParam[j].param,keyword))
      {
        keyExists = TRUE;
        if (daqTbl[i].active) activeId = i;
        *rec = daqTbl[i].rec;
        *param  = j;
        getDataFilename(i,yy,mm);
        if (fileExists(i,&mtime)) 
        {
          if (mtime > tgt) return i;
          if (daqTbl[i].active) return i;
          best = i;
        }
      } 
    }
  }
  if (best >= 0) return best;
  return keyExists ? activeId : -illegal_property;
}
time_t targetTime = 0;
time_t targetTimeReplacement = 0;
BYTE *targetValue= NULL;
time_t valueReplacementStartTime = 0;

int trunclen = 0;
float scaleCorrection = 1;
float offsetCorrection = 0;
time_t correctionRangeStart = 0;
time_t correctionRangeEnd = 0;
int correctionIndex = -1;
void applyCorrections(void *dptr,short fmt,int dump)
{
  short sval;
  int ival;
  float fval;
  double dval;
  if (scaleCorrection == 1 && offsetCorrection == 0) return;
  switch (fmt)
  {
    case CF_INT16:
      sval = *(short *)dptr;
      if (dump) printf("%d -> ",sval);
      sval *= (short)scaleCorrection;
      sval += (short)offsetCorrection;
      *(short *)dptr = sval;
      break;
    case CF_INT32:
      ival = *(int *)dptr;
      if (dump) printf("%d -> ",ival);
      ival *= (int)scaleCorrection;
      ival += (int)offsetCorrection;
      *(int *)dptr = ival;
      break;
    case CF_FLOAT:
      fval = *(float *)dptr;
      if (dump) printf("%g -> ",fval);
      fval *= (float)scaleCorrection;
      fval += (float)offsetCorrection;
      *(float *)dptr = fval;
      break;
    case CF_DOUBLE:
      dval = *(double *)dptr;
      if (dump) printf("%g -> ",dval);
      dval *= (double)scaleCorrection;
      dval += (double)offsetCorrection;
      *(double *)dptr = dval;
      break;
    default:
      break;
  }
  return;
}
int archive_heartbeat = 900; // standard heartbeat of 15 minutes (900 seconds)
int insert_heartbeats = FALSE,insert_target = FALSE; // flags
int assignTargetToStoredData = FALSE; // flag for retro-fitting e.g. NAMES
int OpenDbFile(DAQRecord *daqr)
{
  char filenm[256],afile[32];
  int hDataFile = -1;

retryOpen:
  strncpy(afile,daqr->fname,16);
# ifdef WIN32
# define O_readmode O_RDONLY|O_BINARY
  sprintf(filenm,"%s\\%04d\\%02d\\%s",
    ArchiveDatapath,daqr->archYear,daqr->archMonth,afile);
# else
# define O_readmode O_RDONLY
  sprintf(filenm,"%s/%04d/%02d/%s",
    ArchiveDatapath,daqr->archYear,daqr->archMonth,afile);
# endif
  hDataFile = open(filenm,O_readmode);
  return hDataFile;
}
int scan(int n)
{
  int hFile;
  long timebin,pstart,number = 0, cnt = 0, dump = FALSE, skip = FALSE, inrange = TRUE;
  SINT32 timestamp, lasttimestamp = 0, sysstamp = 0, usrstamp = 0;
  double dtimestamp = 0, datatime;
  char s[256],c;
  BYTE hdr[HISTORY_HDR_SIZE], lasthdr[HISTORY_HDR_SIZE];
  DTYPE d;
  unsigned char dummy;
  int fmtsize,i,p,useMinimalStorage,targetTimeMatched=FALSE;
  int recsize, hdrsiz, dsiz;
  time_t monthtime;
  struct tm *lt;
  FILE *fpout=NULL;
  DAQRecord *daqr = &daqTbl[n];

  lt = localtime((time_t *)&startuptime);
retryFileName:
  getDataFilename(n,lt->tm_year,lt->tm_mon+1);
  useMinimalStorage = daqTbl[n].useMinimalStorage;
  if (daqTbl[n].fmt == CF_STRUCT) trunclen = 0;
  if (getFormatHeaderSize(daqTbl[n].fmt) != 0) trunclen = 0;
  if (trunclen > 0)
  {
    fmtsize = fmtsizeof(daqTbl[n].fmt);
    trunclen *= fmtsize;
  }
  if (usepoi)
  {
    daqTbl[n].fname[0] = 'p';
    daqTbl[n].fname[1] = 'i';
  }
  if ((hFile=OpenDbFile(&daqTbl[n])) == -1) 
  {
    if (!useMinimalStorage)
    {
      daqTbl[n].useMinimalStorage = TRUE;
      goto retryFileName;
    }
    goto err;
  }
  if (!scanonly)
  {
    sprintf(s,"%s",daqr->fname);
    if (addext) strcat(s,"-fix");
    if ((fpout=fopen(s,"w")) == NULL) 
    {
      printf("can't open %s for writing\n",s);
      exit(1);
    }
  }
  hdrsiz = getRecordPrefixLength(useMinimalStorage);
  recsize = daqr->reclen + hdrsiz;
  number = 0; cnt = 0;
  lseek(hFile,0,SEEK_SET);
 
  lt->tm_mday = 1; lt->tm_hour = 0; lt->tm_min = 0; lt->tm_sec = 0;
  monthtime = mktime(lt);
  d.dFormat = daqr->fmt;
  d.dArrayLength = 1;
  while(read(hFile,hdr,hdrsiz) == hdrsiz)
  {
    skip = FALSE;
    if (!useMinimalStorage) 
    { 
      dtimestamp = *(double *)hdr;
      timestamp = (SINT32)dtimestamp;
      sysstamp = *(SINT32 *)&hdr[8];
      usrstamp = *(SINT32 *)&hdr[12];
    }
    else 
    { 
      timestamp = *(SINT32 *)hdr;
      dtimestamp = (double)timestamp; 
    }
check_timestamp:
    if (timestamp < monthtime)
    { // read an earlier (or corrupted?) timestamp
      if (timestamp > monthtime - 86400 * 6 * 30)
      { // something from last half year ...
        if (cnt == 0 && insert_heartbeats)
        { // first one: correct the timestamp
          timestamp = monthtime;
          dtimestamp = (double)timestamp;
          if (dumpoutput) printf("correct initial timestamp to %s\n",GetDataTimeString(dtimestamp,FALSE));
          if (!useMinimalStorage)
            *(double *)hdr = (double)timestamp;
          else
            *(SINT32 *)hdr = timestamp;
        }
        else
        { // skip 
          if (dumpoutput) printf("skip over timestamp %s\n",GetDataTimeString(dtimestamp,FALSE));
          if (read(hFile,traceDataBuffer,daqr->reclen) != daqr->reclen) goto err;
          continue;
        }
      }
      else
      {
        printf("read timestamp %d : %s\n",timestamp,GetDataTimeString(dtimestamp,FALSE));
        if (bailout) exit(1);
        read(hFile,&dummy,1);
        read(hFile,&timestamp,4);
        printf("corrected timestamp : %d : %s\n",timestamp,GetDataTimeString(dtimestamp,FALSE));
        goto check_timestamp;
      }
    }
    if (timestamp < lasttimestamp)
    { // skip over (how did this get in here?)
      if (dumpoutput) printf("skip over out-of-oder timestamp %s\n",GetDataTimeString(dtimestamp,FALSE));
      if (read(hFile,traceDataBuffer,daqr->reclen) != daqr->reclen) goto err;
      continue;  
    }
    if (timestamp == (SINT32)targetTime)
    { // match on a seconds granularilty basis
      targetTimeMatched =  TRUE;
      if (targetValue == NULL && targetTimeReplacement > 0)
      { // replace timestamp in situ
        printf("Replacing timestamp %d with %ld\n",timestamp,targetTimeReplacement);
        timestamp = (SINT32)targetTimeReplacement;
        dtimestamp = (double)timestamp;
        if (!useMinimalStorage)
          *(double *)hdr = (double)timestamp;
        else
          *(SINT32 *)hdr = timestamp;
      }
	    // TODO: find a way to replace values (single values are easy, arrays not so easy ...
	  }
    if (insert_heartbeats)
    {
      if (cnt == 0 && timestamp > monthtime + archive_heartbeat)
      { // need value from previous month
        int hPrevFile;
        struct tm *plt;
        plt = localtime((time_t*)&lastmonthtime);
        getDataFilename(n, plt->tm_year, plt->tm_mon + 1);

        if ((hPrevFile=OpenDbFile(&daqTbl[n])) != -1)
        {
          int fsiz = lseek(hPrevFile, 0, SEEK_END);
          int p = fsiz - daqr->reclen;
          if (lseek(hPrevFile, p, SEEK_SET) > 0)
          {
            if (dumpoutput) printf("time : %s\n", GetDataTimeString(dtimestamp, FALSE));
            if (read(hPrevFile, traceDataBuffer, daqr->reclen) == daqr->reclen)
            {
              memcpy(lasthdr, hdr, hdrsiz);
              if (dumpoutput) printf("correct initial timestamp to %s (insert data from previous month)\n", GetDataTimeString((double)monthtime, FALSE));
              if (!useMinimalStorage)
                *(double*)lasthdr = (double)monthtime;
              else
                *(SINT32*)lasthdr = monthtime;
              if (!scanonly)
              { // write this data set at the beginning
                if (fwrite(lasthdr, hdrsiz, 1, fpout) != 1)
                {
                  printf("can't write timestamp header to %s\n", s);
                  exit(1);
                }
                if (fwrite(traceDataBuffer, daqr->reclen, 1, fpout) != 1)
                {
                  printf("can't write record to %s\n", s);
                  exit(1);
                }
              }
              lasttimestamp = monthtime;
            }
          }
          close(hPrevFile);
        }
      }
      while (lasttimestamp > 0 && timestamp > lasttimestamp + archive_heartbeat)
      {
        lasttimestamp += archive_heartbeat;
        if (dumpoutput) printf("insert data at time : %s\n",GetDataTimeString((double)lasttimestamp,FALSE));
        if (!scanonly)
        { // insert the last read data value at each heartbeat ...
          if (!useMinimalStorage)
	          *(double *)lasthdr = (double)lasttimestamp;
          else
	          *(SINT32 *)lasthdr = lasttimestamp;
          if (fwrite(lasthdr,hdrsiz,1,fpout) != 1)
          {
	          printf("can't write timestamp header to %s\n",s);
	          exit(1);
          }
          if (fwrite(traceDataBuffer,daqr->reclen,1,fpout) != 1)
          {
	          printf("can't write record to %s\n",s);
	          exit(1);
          }
        }
      }
      memcpy(lasthdr,hdr,hdrsiz);
    }
    lasttimestamp = timestamp;
    if (dumpoutput) printf("time : %s\n",GetDataTimeString(dtimestamp,FALSE));
    if (read(hFile,traceDataBuffer,daqr->reclen) != daqr->reclen) goto err;
    if (assignTargetToStoredData && targetTimeMatched)
    {
      number = daqr->reclen;
      goto err;
    }
    if (timestamp >= valueReplacementStartTime && timestamp < targetTime && targetDataBuffer != NULL)
    { // we're in range !
      if (timestamp != valueReplacementStartTime && insert_target)
      { // write the desired data at the replacement start time ...
        if (!useMinimalStorage)
          *(double*)lasthdr = (double)valueReplacementStartTime;
        else
          *(SINT32*)lasthdr = valueReplacementStartTime;
        if (!scanonly)
        { // write this data set at the beginning
          if (fwrite(lasthdr, hdrsiz, 1, fpout) != 1)
          {
            printf("can't write timestamp header to %s\n", s);
            exit(1);
          }
          if (fwrite(targetDataBuffer, daqr->reclen, 1, fpout) != 1)
          {
            printf("can't write record to %s\n", s);
            exit(1);
          }
        }       
      }
      insert_target = FALSE;
      memcpy(traceDataBuffer,targetDataBuffer,daqr->reclen);
      if (dumpoutput) printf("data : replaced with values at %s\n",GetDataTimeString((double)targetTime,FALSE));
      goto out;
    }
      inrange = (timestamp < correctionRangeStart || timestamp > correctionRangeEnd) ? FALSE : TRUE;
    for (i=0; i<daqr->nparam; i++)
    {
      if (dumpoutput) printf("param : %s\ndata : ", daqr->daqParam[i].param);
      d.data.bptr = traceDataBuffer + daqr->daqParam[i].offset;
      fmtsize = fmtsizeof(daqr->daqParam[i].fmt);
      for (p=0; p<daqr->daqParam[i].size; p++)
      {
        dump = (dumpoutput && (p < outputlen)) ? TRUE : FALSE;
        skip = !inrange;
        if (correctionIndex >= 0 && daqr->daqParam[i].size > 1 && p != correctionIndex)
        {
          skip = TRUE;
        }
        switch (daqr->daqParam[i].fmt)
        {
          case CF_INT16: 
            if (!skip) applyCorrections(&d.data.sptr[p],daqr->daqParam[i].fmt,dump);
            if (dump) printf("%d ",d.data.sptr[p]);
            break;
          case CF_INT32: 
            if (!skip) applyCorrections(&d.data.lptr[p],daqr->daqParam[i].fmt,dump);
            if (dump) printf("%d ",d.data.lptr[p]);
            break;
          case CF_FLOAT:
            if (!skip) applyCorrections(&d.data.fptr[p],daqr->daqParam[i].fmt,dump);
            if (dump) printf("%g ",d.data.fptr[p]);
            break;
          case CF_DOUBLE:
            if (!skip) applyCorrections(&d.data.dptr[p],daqr->daqParam[i].fmt,dump);
            if (dump) printf("%g ",d.data.dptr[p]);
            break;
          case CF_NAME16:
            if (dump) printf("%.16s ",&d.data.cptr[p*16]);
            break;
          case CF_NAME32:
            if (dump) printf("%.32s ",&d.data.cptr[p*32]);
            break;
          case CF_NAME64:
            if (dump) printf("%.64s ",&d.data.cptr[p*64]);
            break;
          default: 
            if (trunclen > 0)
            {
              if (dump) printf("format %.16s data will be truncated from %d to %d entries\n",itof(daqr->daqParam[i].fmt),daqr->reclen,trunclen);
              p = daqr->daqParam[i].size;
              i = daqr->nparam;
              break;            
            }
            if (pruneLength > 1)
            {
              if (dump) printf("format %.16s data will be pruned by a factor of %d\n", itof(daqr->daqParam[i].fmt), pruneLength);
              p = daqr->daqParam[i].size;
              i = daqr->nparam;
              break;
            }
            dsiz = 256;
            if (GetValuesArrayAsString(&d,s,&dsiz,NULL) != 0)
            {
              if (dump) printf("format %d unexpected\n",daqr->daqParam[i].fmt);
              goto err;
            }
            if (dump) printf("%.256s ",s);
            d.data.bptr += fmtsize;
            break;
        }
      }
      if (dumpoutput) printf("\n");
    }
    if (rmvrange)
    {
      if (timestamp >= rangestart && timestamp <= rangestop)
      {
        printf("remove stored data at time %s\n",GetDataTimeString(dtimestamp,FALSE));
        continue;      
      }
    }
out:
    cnt++;
    if (!scanonly)
    {
      int reclen = trunclen > 0 ? trunclen : daqr->reclen;
      if (pruneLength > 1 && (cnt-1)%pruneLength) continue;
      if (fwrite(hdr,hdrsiz,1,fpout) != 1)
      {
        printf("can't write timestamp header to %s\n",s);
        exit(1);
      }
      if (fwrite(traceDataBuffer,reclen,1,fpout) != 1)
      {
        printf("can't write record to %s\n",s);
        exit(1);
      }
    }
    number++;
  }
  if (insert_heartbeats)
  { // then go to the end of the month ...
    time_t tnow = time(NULL);
    lt->tm_mday = nDaysInMonth(lt->tm_year,lt->tm_mon);
    lt->tm_hour = 23;
    lt->tm_min = 59; lt->tm_sec = 59;
    monthtime = mktime(lt);
    if (monthtime > tnow) monthtime = tnow;
    while (lasttimestamp > 0 && monthtime > lasttimestamp + archive_heartbeat)
    {
	    lasttimestamp += archive_heartbeat;
	    if (dumpoutput) printf("insert data at time : %s\n",GetDataTimeString((double)lasttimestamp,FALSE));
	    if (!scanonly)
	    { // insert the last read data value at each heartbeat ...
        if (!useMinimalStorage)
	        *(double *)lasthdr = (double)lasttimestamp;
        else
	        *(SINT32 *)lasthdr = lasttimestamp;
        if (fwrite(lasthdr,hdrsiz,1,fpout) != 1)
        {
	        printf("can't write timestamp header to %s\n",s);
	        exit(1);
        }
        if (fwrite(traceDataBuffer,daqr->reclen,1,fpout) != 1)
        {
	        printf("can't write record to %s\n",s);
	        exit(1);
        }
 	    }
    }
  }  
err: 
  if (hFile != -1) close(hFile);
  if (!scanonly && fpout != NULL) fclose(fpout);
  if (assignTargetToStoredData && !targetTimeMatched) return 0;
  return number;
}
time_t getTimeFromString(char *str)
{
  struct tm tms;
  tms.tm_mday = atoi(str);
  tms.tm_mon = atoi(&str[3]) - 1;
  tms.tm_year = 100 + atoi(&str[6]);
  tms.tm_hour = atoi(&str[9]);
  tms.tm_min = atoi(&str[12]);
  tms.tm_sec = atoi(&str[15]);
  tms.tm_isdst = -1;
  return mktime(&tms);
}
char *usageTxt[] = {
  "\nscans and massages the targeted archive data file\n",
  "dbscan /f=<database file name> (/k=<archive keyword> or /r=<record>)", 
  "[/y=<year> /m=<month> /x= or /scale=<scale correction> /o=<offset correction>", 
  " /d=<start::stop - deletes all records found in range>",
  " /e=<target time::replacment time - corrects a stored timestamp (if found) to the given replacement value>",
  " /vr=<replacement start time::replacment time - replaces stored data with values collected at the target time (e.g. a names list)>",
  " /range=<range start time::range stop time - only apply corrections within the range given>",
  " /heartbeats=<TRUE> or <heartbeat> - insert missing heartbeat records; value > 60 will override the default of 900 seconds",  
  " /index=specific array index - only apply corrections (scale, offset) to specific array index",  
  " /scan=<TRUE: scan only (don't create 'fix' file)>",
  " /v=<TRUE: (verbose) dump scanned contents> /l=<dump array length>",
  " /b=<TRUE: bailout on error> /a=<TRUE: add '-fix' extension>",
  " /n=<new record length> /poi=<TRUE: scan POI file>",
  " /prune=<prune data skip raster>]",
  " /poi=<scan points-of-interest data>]",
  "\ne.g. dbscan /f=datapetr.csv /r=1791 /y=2018 /m=10 /p=10",
  "\n\tscans record 1791 of the 'datapetr.csv' database from October 2018 and prunes the",
  "\tdata by taking every 10th stored value\n",
  "\ne.g. dbscan /f=datapetr.csv /r=1700 /y=2018 /m=10 /vr=01.08.2018_10:00:00::14.09.2018_17:45:03",
  "\n\tscans record 1700 of the 'datapetr.csv' database from October 2018 and replaces all records",
  "\tbeginning at 01.08.2018 with the record found at 14.08.2018 (e.g. correcting a names list",
};
#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))

void dumpUsage(int level)
{
  int i;
  for (i = 0; i<nUsageLines; i++)
  {
    printf(i == 1 ? "Usage : " : "        ");
    printf("%s\n", usageTxt[i]);
  }
  printf("\n");
  exit(level);
}
int main(int argc, char *argv[])
{
  int i,cc,yy,mm,rn=-1,p,slen;
  char *c,*ce,scratch[256],kwd[64];
  time_t tnow = time(NULL);
  struct tm *lt;
  lt = localtime(&tnow);
  yy = lt->tm_year + 1900;
  mm = lt->tm_mon + 1;
  if (argc < 3) dumpUsage(1);
  for (i=1; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 'f': // database file name
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(dbfile,c,60);
          if (strchr(dbfile,'.') == NULL) strcat(dbfile,".csv");
        }
        continue;
      case 'k': // archive keyword
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(kwd,c,64);
        }
        continue;
      case 'o': // offset correction
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,64);
          offsetCorrection = (float)strtod(scratch,NULL);
          printf("shifting stored data by %g\n", offsetCorrection);
        }
        continue;
      case 'e': // timestamp or value exchange
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,248);
          if ((c=strstr(scratch,"::")) != NULL)
          { // timestamp replacement given
            *c = 0; c += 2;
            targetTime = (time_t)GetTimeStampFromString(scratch);
            targetTimeReplacement = (time_t)GetTimeStampFromString(c);
            printf("correct time %.26s (%ld) to %.26s (%ld)\n", scratch, targetTime, c, targetTimeReplacement);
            continue;
          }
          if ((c=strstr(scratch,"+")) != NULL)
          { // timestamp and value replacement given
            double v = 0.0;
            targetTime = (time_t)GetTimeStampFromString(scratch);
            targetValue = (BYTE *)&v;
            printf("correct time %.26s (%ld) to value %g (not yet implemented)\n", scratch, targetTime, *(double *)targetValue);
            targetTime = 0;
            continue;
          }
          printf("unable to parse data exchange %s\n", scratch);
        }
        continue;
      case 'n': // new record length
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,64);
          trunclen = atoi(c);
          printf("truncating stored data array to length %d \n", trunclen);
        }
        continue;
      case 'x': // scale correction (same as /scale= )
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,64);
          scaleCorrection = (float)strtod(scratch,NULL);
          if (scaleCorrection == 0)
          {
            printf("invalid scale factor: multiplying data by 0 is not allowed\n");
            exit(1);
          }
          printf("scaling stored data by %g\n",scaleCorrection);
        }
        continue;
      case 'd': // delete records over a time range
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,128);
          if ((c=strstr(scratch,"::")) == NULL)
          {
            printf("the delete records option must specify a range in the form dd.mm.yy_hh:mm:ss::dd.mm.yy_hh:mm:ss\n");
            exit(1);
          }
          *c = 0; c += 2;
          rangestart = getTimeFromString(scratch);
          rangestop = getTimeFromString(c);
          rmvrange = TRUE;
          printf("delete stored records from %s (%ld) to %s (%ld)\n", 
            scratch,rangestart,c,rangestop);
        }
        continue;
      case 's':  // scan only or scale
        if ((ce=strchr(c,'=')) != NULL)
        {
          *ce++ = 0;
          slen = (int)strlen(c);
          if (!strnicmp(c,"scan",slen))
          { // scan only
            strncpy(scratch,ce,64);
            if (stricmp(scratch,"TRUE") == 0 || stricmp(scratch,"YES") == 0) scanonly = TRUE;
          }
          else if (!strnicmp(c,"scale",slen))
          { // value replacement
            strncpy(scratch,ce,64);
            scaleCorrection = (float)strtod(scratch,NULL);
            if (scaleCorrection == 0)
            {
              printf("invalid scale factor: multiplying data by 0 is not allowed\n");
              exit(1);
            }
            printf("scaling stored data by %g\n",scaleCorrection);
          }
        }
        continue;
      case 'a': // add '-fix' extension to new file name
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,64);
          if (stricmp(scratch,"TRUE") == 0 || stricmp(scratch,"YES") == 0) addext = TRUE;
        }
        continue;
      case 'i':  // insert heartbeats or index of array
        if ((ce=strchr(c,'=')) != NULL)
        {
          int v = 0;
          *ce++ = 0;
          slen = (int)strlen(c);
          if (!strnicmp(c,"insert",slen))
          { // insert heartbeats (e.g. /i=TRUE)
            strncpy(scratch,ce,64);
            if (stricmp(scratch,"TRUE") == 0 || stricmp(scratch,"YES") == 0) insert_heartbeats = TRUE;
            v = atoi(scratch);
            if (v > 60)
            {
              insert_heartbeats = TRUE;
              archive_heartbeat = v;
            }
            printf("insert missing heartbeats : %s every %d seconds\n", (insert_heartbeats ? "TRUE" : "FALSE"), archive_heartbeat);
          }
          else if (!strnicmp(c,"index",slen))
          { // specific array index to use for corrections (e.g. /index=2)
            strncpy(scratch,ce,64);
            correctionIndex = atoi(scratch);
          }
        }
        continue;
      case 'v': // verbose or value replacement 
        if ((ce=strchr(c,'=')) != NULL)
        {
          *ce++ = 0;
          slen = (int)strlen(c);
          if (!strnicmp(c,"verbose",slen))
          { // verbose
            strncpy(scratch,ce,64);
            if (stricmp(scratch,"FALSE") == 0 || stricmp(scratch,"NO") == 0) dumpoutput = FALSE;
          }
          else if (!strnicmp(c,"vr",2))
          { // value replacement
            strncpy(scratch,ce,248);
            if ((ce=strstr(scratch,"::")) != NULL)
            { // timestamp replacement given
              *ce = 0; ce += 2;
              valueReplacementStartTime = (time_t)GetTimeStampFromString(scratch);
              targetTime = (time_t)GetTimeStampFromString(ce);
              assignTargetToStoredData = TRUE;
              printf("retro fit values to those at %.26s (%ld)\n", scratch, targetTime);
              continue;
            }
            printf("unable to parse data fill %s\n", scratch);
          }
        }
        continue;
      case 'l':  // dump array length (applies to screen dump)
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,64);
          outputlen = atoi(scratch);
        }
        continue;
      case 'y': // target year 
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,64);
          yy = atoi(scratch);
          if (yy < 100) yy += 2000;
        }
        continue;
      case 'm': // target month
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,64);
          mm = atoi(scratch);
        }
        continue;
      case 'b':  // bailout on error
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,64);
          if (stricmp(scratch,"FALSE") == 0 || stricmp(scratch,"NO") == 0) bailout = FALSE;
          printf("Stop scanning on error : %s\n", (bailout ? "TRUE" : "FALSE"));
        }
        continue;
      case 'r': // record number (alternative to keyword) or range 
        if ((ce=strchr(c,'=')) != NULL)
        {
          *ce++ = 0;
          slen = (int)strlen(c);
          strncpy(scratch,c,64);
          if (!strnicmp(c,"record",slen))
          { // record number
            strncpy(scratch,ce,64);
            rn = atoi(scratch);
          }
          else if (!strnicmp(c,"range",slen))
          { // range
            strncpy(scratch,ce,248);
            if ((ce=strstr(scratch,"::")) != NULL)
            { // start and stop given
              *ce = 0; ce += 2;
              correctionRangeStart = (time_t)GetTimeStampFromString(scratch);
              correctionRangeEnd = (time_t)GetTimeStampFromString(ce);
              printf("apply any corrections only from %.26s to %.26s\n", scratch, ce);
            }
            else
            {
              correctionRangeStart = (time_t)GetTimeStampFromString(scratch);
              correctionRangeEnd = time(NULL);
              printf("apply any corrections only from %.26s to now\n", scratch);
            }
          }
        }
        continue;
      case 'p':  // pruning length or scan points of interest 
      {
        if ((ce=strchr(c,'=')) != NULL)
        {
          int v = 0;
          *ce++ = 0;
          slen = (int)strlen(c);
          if (!strnicmp(c,"prune",slen))
          { // prune
            pruneLength = atoi(ce);
            printf("prune stored data by taking every %dth data point\n", pruneLength);
          }
          else if (!strnicmp(c,"poi",slen))
          { // scan points of interest
            strncpy(scratch,ce,64);
            if (stricmp(scratch,"TRUE") == 0 || stricmp(scratch,"YES") == 0) usepoi = TRUE; 
            printf("Scan points-of-interest data file : %s\n", (usepoi ? "TRUE" : "FALSE"));			  
          }
        }
        continue;
      }
      case '?':
        dumpUsage(0);
        exit(0);
      case 'h': // help
        if ((ce=strchr(c,'=')) != NULL)
        {
          int v = 0;
          *ce++ = 0;
          slen = (int)strlen(c);
          if (!strnicmp(c,"help",slen))
          { // help
            dumpUsage(0);
            exit(0);
          }
          else if (!strnicmp(c,"heartbeats",slen))
          { // insert heartbeats (e.g. /heartbeats=TRUE)
            strncpy(scratch,ce,64);
            if (stricmp(scratch,"TRUE") == 0 || stricmp(scratch,"YES") == 0) insert_heartbeats = TRUE;
            v = atoi(scratch);
            if (v > 60)
            {
              insert_heartbeats = TRUE;
              archive_heartbeat = v;
            }
            printf("insert missing heartbeats : %s every %d seconds\n", (insert_heartbeats ? "TRUE" : "FALSE"), archive_heartbeat);
          }
        }
        continue;
      default:
        continue;
    }
  }
  lt->tm_year = yy - 1900;
  lt->tm_mon = mm - 1;
  lt->tm_mday = 1;
  lt->tm_hour = 0;
  lt->tm_min = 0;
  lt->tm_sec = 1;
  lt->tm_isdst = -1;
  startuptime = mktime(lt);
  if (lt->tm_mon > 0)
  {
    lt->tm_mon--;
  }
  else
  {
    lt->tm_mon = 11;
    lt->tm_year--;
  }
  lastmonthtime = mktime(lt);
  initScanner();

  if (rn < 0)
  { // get record number from input keyword
    if ((i=getArchivedEntryFromKeywordAndTime(kwd,&rn,&p,startuptime)) < 0) 
    {
      printf("record %d not found in database\n",rn);
      exit(2);
    }
  }
  else
  {
    if ((i=getEntryFromRectyp(rn)) < 0)
    {
      printf("record %d not found in database\n",rn);
      exit(2);
    }
  }
  printf("scanning record %s (rec. no. %d), VERBOSE is %s: type 'x' to exit, any other key to continue ...\n",daqTbl[i].prp,rn,dumpoutput?"ON":"OFF");
  if (getchar() == 'x') exit(0);
  if (correctionRangeEnd == 0) correctionRangeEnd = time(NULL);
  if (assignTargetToStoredData && targetTime > 0)
  {
    time_t interim_t = startuptime;
    int interim_scanonly = scanonly;
    scanonly = TRUE;
    startuptime = targetTime;
    targetDataBufferSize = scan(i);
    startuptime = interim_t;
    scanonly = interim_scanonly;    
    targetDataBuffer = (BYTE *)malloc(targetDataBufferSize);
    if (targetDataBuffer == NULL)
    {
      printf("cannot allocate %d bytes to store target data!\n",targetDataBufferSize);
      exit(3);
    }
    memcpy(targetDataBuffer,traceDataBuffer,targetDataBufferSize);
    assignTargetToStoredData = FALSE;
    insert_target = TRUE;
  }
  
  scan(i);

  return 0;
}
